/*
 * OCSP responder
 * by Massimiliano Pala (madwolf@openca.org)
 * OpenCA project 2001
 *
 * Copyright (c) 2001 The OpenCA Project.  All rights reserved.
 *
 * ====================================================================
 *
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).  This product includes software written by Tim
 * Hudson (tjh@cryptsoft.com).
 *
 */

#include "http_client.h"
#include "support.h"

extern OCSPD_CONFIG *ocspd_conf;

/* Functions */
BIO *ocspd_http_connect( URL *url ) {

	BIO *cbio = NULL;
	char buf[1024];

	if( (!url) || (!url->addr) ) {
		syslog( LOG_ERR, "ERROR: Missing address for HTTP connect");
		return NULL;
	}

	/* We build the "host:port" string to be used */
	memset(buf, '\x0', sizeof(buf));
        sprintf( buf, "%s:%d", url->addr, url->port );

	if((cbio = BIO_new_connect(buf)) == NULL ) {
		syslog(LOG_ERR, "ERROR: cannot connect to %s", url->addr);
		return NULL;
	}

	if (BIO_do_connect(cbio) <= 0) {
		syslog(LOG_ERR, "ERROR: cannot connect to %s", url->addr);
		BIO_free_all(cbio);
		return NULL;
	}

	return cbio;
}

BUF_MEM *ocspd_http_get_data ( BIO *in, ssize_t max_size ) {

	BUF_MEM *buf = NULL;
	unsigned long fullsize = 0;
	unsigned long newsize  = 0;

	if( !in )
		return NULL;

	/* get subject name from bio using recommended OpenSSL template */
	buf = BUF_MEM_new();
	for (;;) {
		if ((buf == NULL) || (!BUF_MEM_grow(buf, fullsize+512))) {
			return NULL;
		}

		newsize   = BIO_read(in, &(buf->data[fullsize]), 512);
		fullsize += newsize;
		if (newsize == 0) break;

		if (newsize < 0) {
			/* Error Reading from buf */
			BUF_MEM_free( buf );
			return NULL;
		}

		if( (max_size > 0) && (fullsize > max_size)) {
			/* Max Reading size exceeded */
			syslog( LOG_ERR, 
				"ERROR::HTTP::Read::Max read size exceeded "
				" [ %ld ]", (long) max_size );
			BUF_MEM_free( buf );
			return NULL;
		}
	}

	/* Make sure the buf ends with a \x0 value */
	buf->data[fullsize] = '\x0';

	return buf;
}

BUF_MEM *ocspd_http_get ( URL *url, unsigned long max_size, char *version ) {
	BIO *in = NULL;
	BUF_MEM *buf = NULL;
	char *def_version = "1.1";

	char *get_headers[] = {
		"GET %s HTTP/%s\r\n",
		"Host: %s\r\n",
		NULL
	};

	in = ocspd_http_connect( url );
	if( !in ) return NULL;

	if( ocspd_conf->verbose ) {
		syslog( LOG_INFO, "INFO::Connected to HTTP server");
		// syslog( LOG_INFO, get_s, url->path );
	}

	if( version == NULL ) {
		version = def_version;
	}

	/* Print the "GET" command to the server and wait for the
	   answer */
	BIO_printf(in, get_headers[0], url->path, version);
	if( ocspd_conf->debug ) {
		fprintf( stderr, get_headers[0], url->path, version );
	}

	/* For HTTP/1.1 let's specify the server name as well */
	if( strncmp_nocase("1.1", version, 3) == 0 ) {
		if( ocspd_conf->debug ) {
			fprintf( stderr, get_headers[1], url->addr );
		}
		BIO_printf( in, get_headers[1], url->addr );
	}

	/* Print EMPTY line for the end of the HEADERS */
	BIO_printf(in, "\r\n");
	if( ocspd_conf->debug ) {
		fprintf( stderr, "\r\n");
	}

	/* Get HTTP data */
	if((buf = ocspd_http_get_data ( in, max_size )) == NULL ) {
		syslog( LOG_ERR, "ERROR -- can not read http data.");
	}

	if( ocspd_conf->debug ) {
		fprintf( stderr, "---BEGIN RECEIVED DATA---\n" );
		fprintf( stderr, "%s\n---END OF RECEIVED DATA---\n", buf->data);
	}

	/* Release the socket */
	BIO_free_all (in);
	
	return buf;
}

int ocspd_parse_http_headers ( BIO *in ) {

	int head_lines = 0;
	char linebuf[1024];

	head_lines = 0;
	do {
		int read_code = 0;

		memset(linebuf, '\x0', sizeof linebuf);
        	if((read_code = BIO_gets(in, linebuf, sizeof(linebuf))) < 0) {
			syslog( LOG_ERR, "HTTP - Error (%d) retrieving data"
					" (head_lines=%d)", 
						read_code, head_lines );
			return 0;
        	}

		/* Let's check if the first line has good code */
		if ( ( head_lines == 0 ) && 
			( strncmp_nocase( &(linebuf[9]), "200", 3) != 0 )) {
				/* Got an error - probably not found (?) */
				syslog( LOG_ERR, "HTTP - Error retrieving data"
					" (%s)", linebuf );
				return 0;
		} else {
			head_lines++;
		}
	}
	while( (linebuf[0] != '\r') && (linebuf[0] != '\n') );

	if( ocspd_conf->verbose ) {
		syslog( LOG_INFO, "INFO::Parsed HTTP HEADERS");
	}

	return 1;

}

